-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("tooltips", "AceHook-3.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, tonumber = pairs, tonumber
local format, ceil, match, GetAchievementCriteriaInfo = string.format, math.ceil, string.match, GetAchievementCriteriaInfo

-----------------
--- Variables ---
-----------------
local origOnShow, origOnTooltipSetItem

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		mobTooltip = true,
		itemTooltip = true,
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("tooltips", defaults, "Default")
	self:SetEnabledState(core:getModuleEnabled("tooltips"))
	core.moduleState["tooltips"] = core:getModuleEnabled("tooltips")
end

---------------------------
--- Main Module Program ---
---------------------------
local function OnTooltipSetUnit(tooltip, ...)
	tooltip = tooltip or GameTooltip
	if not tooltip then return end
	local name, unit = tooltip:GetUnit()
	if not unit then return end
	local ttmod
	if module.db.profile.mobTooltip then
		if core.questsMobs[name] then
			for uid, qmData in pairs(core.questsMobs[name].quests) do
				tooltip:AddDoubleLine(core.quests[uid].title, core.questsObjectives[uid][qmData].qoStatus)
			end
			ttmod = true
		end
		if core.achievementsMobs[name] then
			for _, data in pairs(core.achievementsMobs[name]) do
				local aid, cid = match(data, "(.+)~(.+)")
				local _, _, _, completed = GetAchievementCriteriaInfo(aid, cid)
				if completed == 1 then
					tooltip:AddLine(core.achievementName[tonumber(aid)], 0, 1, 0)
				else
					tooltip:AddLine(core.achievementName[tonumber(aid)], 1, 0, 0)
				end
				tooltip:AddTexture(core.achievementIcon[tonumber(aid)])
			end
			ttmod = true
		end
	end
	if ttmod then tooltip:Show() end
end

local function OnTooltipSetItem(tooltip, ...)
	tooltip = tooltip or GameTooltip
	if not tooltip then return end
	local name = tooltip:GetItem()
	if not name then return end
	local ttmod
	if module.db.profile.itemTooltip then
		if core.questsItems[name]then
			for uid, qiData in pairs(core.questsItems[name].quests) do
				tooltip:AddDoubleLine(core.quests[uid].title, core.questsObjectives[uid][qiData].qoStatus)
			end
			ttmod = true
		end
		if core.achievementsItems[name] then
			for _, data in pairs(core.achievementsItems[name]) do
				local aid, cid = match(data, "(.+)~(.+)")
				local _, _, _, completed = GetAchievementCriteriaInfo(aid, cid)
				if completed == 1 then
					tooltip:AddLine(core.achievementName[tonumber(aid)], 0, 1, 0)
				else
					tooltip:AddLine(core.achievementName[tonumber(aid)], 1, 0, 0)
				end
				tooltip:AddTexture(core.achievementIcon[tonumber(aid)])
			end
			ttmod = true
		end
	end
	if ttmod then tooltip:Show() end
	if origOnTooltipSetItem then return origOnTooltipSetItem(tooltip, ...) end
end

local function OnShow(tooltip, ...)
	tooltip = tooltip or GameTooltip
	if tooltip and tooltip:NumLines() == 1 then
		local name = _G["GameTooltipTextLeft1"]:GetText()
		if not name then return end
		local ttmod
		if module.db.profile.itemTooltip then
			if core.achievementsItems[name] then
				for _, data in pairs(core.achievementsItems[name]) do
					local aid, cid = match(data, "(.+)~(.+)")
					local _, _, _, completed = GetAchievementCriteriaInfo(aid, cid)
					if completed == 1 then
						tooltip:AddLine(core.achievementName[tonumber(aid)], 0, 1, 0)
					else
						tooltip:AddLine(core.achievementName[tonumber(aid)], 1, 0, 0)
					end
					tooltip:AddTexture(core.achievementIcon[tonumber(aid)])
				end
				ttmod = true
			end
		end
		if ttmod then tooltip:Show() end
	end
	if origOnShow then return origOnShow(tooltip, ...) end
end

function module:OnEnable()
	origOnShow = GameTooltip:GetScript("OnShow")
	GameTooltip:SetScript("OnShow", OnShow)
	origOnTooltipSetItem = GameTooltip:GetScript("OnTooltipSetItem")
	GameTooltip:SetScript("OnTooltipSetItem", OnTooltipSetItem)
	GameTooltip:HookScript("OnTooltipSetUnit", OnTooltipSetUnit)
end

function module:OnDisable()
	GameTooltip:SetScript("OnShow", origOnShow)
	GameTooltip:SetScript("OnTooltipSetItem", origOnTooltipSetItem)
end